import base64
import logging
import typing
from collections.abc import Iterable

import arrow
from _pytest.python import Function
from selenium.webdriver.remote.webdriver import WebDriver

from . import constants, storage


class DebugInfo:
    """Representation of debug information."""

    def __init__(self, item: Function, webdriver: WebDriver):
        """Set error log and extract data from extra."""
        self.webdriver: WebDriver = webdriver
        self.test_name = item.name
        self.logger = logging.getLogger(__name__)
        self.screenshot = self._extract_screenshot()
        self.html = self._extract_html()
        self.browser_log = self._extract_browser_log()
        self.url = self._extract_url()

    def _extract_screenshot(self) -> bytes | None:
        try:
            return base64.b64decode(self.webdriver.get_screenshot_as_base64().encode("utf-8"))
        except Exception:
            self.logger.error(msg="Can't extract screenshot from webdriver", exc_info=True)
            return None

    def _extract_html(self) -> bytes | None:
        try:
            return self.webdriver.page_source.encode("utf-8")
        except Exception:
            self.logger.error(msg="Can't extract html page source from webdriver", exc_info=True)
            return None

    def _extract_url(self) -> str:
        try:
            return self.webdriver.current_url
        except Exception:
            self.logger.error(msg="Can't extract url from webdriver", exc_info=True)
            return ""

    def _extract_browser_log(self) -> str:
        logs = []
        try:
            for name in self.webdriver.log_types:
                logs.append(self._format_log(self.webdriver.get_log(name)))  # type: ignore
        except Exception:
            # Sometimes there can be problems reading some logs from the browser here
            # (such as `ProtocolError('Connection broken')`). So we skip them if this happens.

            # Also, this method can't work correctly with Geckodriver (raises `WebDriverException`
            # error) because of the following issue
            # https://github.com/mozilla/geckodriver/issues/284
            self.logger.error(msg="Can't extract browser log", exc_info=True)
            pass
        return "\n".join(logs)

    @staticmethod
    def _format_log(log: Iterable[dict[str, typing.Any]]) -> str:
        """Format logs.

        Copied from pytest-selenium.

        """
        timestamp_format = "%Y-%m-%d %H:%M:%S.%f"
        entries: list[str] = []
        for entry in log:
            timestamp = arrow.get(entry["timestamp"] / 1000.0).strftime(timestamp_format)
            entries.append(f"{timestamp} {entry['level']} - {entry['message']}")
        return "\n".join(entries)

    def generate_debug_comment(
        self,
        file_storage: storage.FileStorage,
        folder: str,
    ) -> str:
        screenshot_url = ""
        if self.screenshot:
            try:
                screenshot_url = file_storage.save_file_obj(
                    content=self.screenshot,
                    filename=f"{folder}/screenshot.png",
                )
            except Exception:
                self.logger.error(msg="Can't save screenshot to storage", exc_info=True)

        html_url = ""
        if self.html:
            try:
                html_url = file_storage.save_file_obj(
                    content=self.html,
                    filename=f"{folder}/html.html",
                )
            except Exception:
                self.logger.error(msg="Can't save HTML to storage", exc_info=True)

        try:
            browser_log_url = file_storage.save_file_obj(
                content=self.browser_log.encode("utf-8"),
                filename=f"{folder}/browser_log.txt",
            )
        except Exception:
            self.logger.error(msg="Can't save browser log to storage", exc_info=True)
            browser_log_url = ""

        return constants.FAILED_TEST_REPORT_TEMPLATE.format(
            url=self.url,
            screenshot_url=screenshot_url,
            html_url=html_url,
            browser_log_url=browser_log_url,
        )



import logging
import os
import pathlib
from typing import cast

import filelock
import pytest
from _pytest.terminal import TerminalReporter
from qase.api_client_v1.exceptions import ApiException
from qase.api_client_v1.models.result_create import ResultCreate
from qase.api_client_v1.models.run import Run

from . import api_client, converter, plugin_exceptions, storage


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add custom args to command line."""
    parser.addoption(
        "--qase-enabled",
        action="store_true",
        default=False,
        help="Store run result in qase",
    )
    parser.addoption(
        "--qase-file-storage",
        default="qase",
        help="Choose file storage to upload debug files",
    )


def pytest_addhooks(pluginmanager: pytest.PytestPluginManager) -> None:
    """Add hooks for plugin."""
    from . import hooks

    pluginmanager.add_hookspecs(hooks)


def _get_file_storage(config: pytest.Config) -> storage.FileStorage | None:
    """Provide file storage via pytest config."""
    file_storage_name: str = config.getoption("--qase-file-storage")
    if file_storage_name.lower() == "none":
        return None

    file_storages: dict[str, storage.FileStorage] = config.hook.pytest_qase_file_storages()

    if file_storage_name not in file_storages:
        logging.getLogger("qase").error(
            "Cannot find registered file storage for "
            f"`{file_storage_name}`, saving to storage is disabled. "
            f"Available storages: {file_storages}",
        )

    return file_storages.get(file_storage_name)


@pytest.hookimpl(trylast=True)
def pytest_qase_file_storages() -> dict[str, storage.FileStorage]:
    """Provide mapping of available file storages for qase debug files."""
    return {
        "qase": storage.QaseFileStorage(
            qase_token=os.environ["QASE_TOKEN"],
            qase_project_code=os.environ["QASE_PROJECT_CODE"],
        ),
    }


@pytest.hookimpl(trylast=True)
def pytest_qase_browser_name(config: pytest.Config) -> str:
    """Try to get browser name from `webdriver` pytest option."""
    return config.getoption("--webdriver")


@pytest.hookimpl(trylast=True)
def pytest_configure(config: pytest.Config) -> None:
    """Configure pytest-qaseio plugin.

    Add `qase` marker for pytest.
    If qase enabled, register qase plugin.
    Get file storage for qase plugin.

    """
    config.addinivalue_line(
        "markers",
        "qase(test_case_url): Mark test for qase",
    )
    qase_enabled = config.getoption("--qase-enabled")
    if not qase_enabled:
        return

    browser_name: str = config.hook.pytest_qase_browser_name(config=config)

    config.pluginmanager.register(
        plugin=QasePlugin(
            browser=browser_name,
            file_storage=_get_file_storage(config),
        ),
        name="qase_plugin",
    )


class QasePlugin:
    """Pytest plugin for reporting tests result to Qase."""

    # We use .pytest-qaseio.lock to lock process, in other words make other workers wait, until
    # worker that locked file will create run and store run's id in .pytest-qaseio. After other
    # workers load run by id from .pytest-qaseio.
    __run_file = pathlib.Path(".pytest-qaseio")
    __run_file_lock = pathlib.Path(".pytest-qaseio.lock")

    def __init__(
        self,
        browser: str,
        file_storage: storage.FileStorage | None,
    ):
        """Save used browser for run's name and folder name."""
        self._client = api_client.QaseClient(
            token=os.environ["QASE_TOKEN"],
            project_code=os.environ["QASE_PROJECT_CODE"],
        )
        self._cases_ids_from_api: list[int] = self._client.load_cases_ids()
        self._current_run: Run | None = None
        self._converter = converter.QaseConverter(
            browser=browser,
            env=os.environ["ENVIRONMENT"],
            project_code=os.environ["QASE_PROJECT_CODE"],
            file_storage=file_storage,
        )

        # Mapping of pytest items ids and case id
        self._tests: dict[str, int | None] = dict()
        # Mapping of case ids and result hash from qase with status
        self._qase_results: dict[int, tuple[str, ResultCreate]] = dict()

    def pytest_sessionstart(self, session: pytest.Session) -> None:
        """Clear previously saved run, prepare lock file."""
        if hasattr(session.config, "workerinput"):
            # Do nothing if it is not master thread
            return
        self.__run_file.unlink(missing_ok=True)
        self.__run_file_lock.touch(exist_ok=True)

    @pytest.hookimpl(trylast=True)
    def pytest_collection_modifyitems(
        self,
        items: list[pytest.Function],
    ) -> None:
        """Create test run in qase."""
        with filelock.FileLock(self.__run_file_lock):
            try:
                run_data, self._tests = self._converter.prepare_run_data(
                    cases_ids_from_api=self._cases_ids_from_api,
                    items=items,
                )

                # Specifying plan allows to create run "from template".
                # New run will contain all cases from plan + cases that
                # specified in tests
                if plan_id := os.getenv("QASE_PLAN_ID"):
                    run_data.plan_id = int(plan_id)

                if environment_id := os.getenv("QASE_ENVIRONMENT_ID"):
                    run_data.environment_id = int(environment_id)

                if qase_url_custom_field_id := os.getenv(
                    "QASE_URL_CUSTOM_FIELD_ID",
                ):
                    run_data.custom_field = {
                        # This should be provided from script that runs test, f.e jenkins script
                        qase_url_custom_field_id: os.getenv("RUN_SOURCE_URL") or "",
                    }

                self._current_run = self._load_run_from_file()
                if self._current_run:
                    return

                self._current_run = self._client.create_run(
                    run_data=run_data,
                )
                with pathlib.Path(self.__run_file).open(mode="w") as lock_file:
                    lock_file.write(str(self._current_run.id))
            except plugin_exceptions.BaseQasePluginException as e:
                pytest.exit(e.message)

    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_runtest_makereport(self, item: pytest.Function):  # noqa: ANN201
        """Represent standard pytest hook on test completion.

        At this hook we will report passed, skipped and failed tests.

        """
        provided_report = yield
        report: pytest.TestReport = provided_report.get_result()
        should_report = not (
            # Passed tests should be reported only on call
            report.passed and report.when in ("setup", "teardown")
        )
        if not should_report:
            return
        case_id = self._tests[item.nodeid]

        # No need to report same passed status,
        # while skipped and failed should be always reported
        if not case_id or (case_id in self._qase_results and report.passed):
            return
        if not self._current_run:
            raise plugin_exceptions.RunNotConfigured()
        try:
            self._qase_results[case_id] = self._client.report_test_results(
                run=self._current_run,
                report_data=self._converter.prepare_report_data(
                    run_id=cast(int, self._current_run.id),
                    case_id=case_id,
                    item=item,
                    report=report,
                ),
            )

        except ApiException as error:
            if report.passed:
                return
            # Qase closes runs, once every case got result. So if try to report any other result,
            # we'll get an error `Test run is not active`.
            terminal_reporter: TerminalReporter = item.config.pluginmanager.get_plugin(
                "terminalreporter",  # type: ignore
            )
            terminal_reporter.ensure_newline()
            terminal_reporter.section(
                f"{error}. "
                f"Seems that Qase closed run, and we are unable to report failed {item.name}",
                sep="=",
            )

    def _load_run_from_file(
        self,
    ) -> Run | None:
        """Load run id and then load it from qase."""
        if not self.__run_file.exists():
            return None
        with pathlib.Path(self.__run_file).open() as lock_file:
            run_id = int(lock_file.read())
            return self._client.get_run(
                run_id=run_id,
            )



import logging
import sys
from typing import cast

from qase.api_client_v1 import configuration as qaseio_config
from qase.api_client_v1.api.cases_api import CasesApi
from qase.api_client_v1.api.results_api import ResultsApi
from qase.api_client_v1.api.runs_api import RunsApi
from qase.api_client_v1.api_client import ApiClient
from qase.api_client_v1.models.id_response_all_of_result import IdResponseAllOfResult
from qase.api_client_v1.models.result_create import ResultCreate
from qase.api_client_v1.models.run import Run
from qase.api_client_v1.models.run_create import RunCreate


class QaseClient:
    """Class for interacting with tests runs in Qase API."""

    def __init__(
        self,
        token: str,
        project_code: str,
    ):
        """Init client."""
        super().__init__()
        self._logger = logging.getLogger("qase")
        self._logger.addHandler(
            logging.StreamHandler(sys.stderr),
        )
        self._client = ApiClient(
            configuration=qaseio_config.Configuration(
                api_key={
                    "TokenAuth": token,
                },
            ),
        )
        self._project_code: str = project_code

    def get_run(
        self,
        run_id: int,
    ) -> Run:
        run_response = RunsApi(self._client).get_run(
            code=self._project_code,
            id=run_id,
        )
        return cast(Run, run_response.result)

    def create_run(
        self,
        run_data: RunCreate,
    ) -> Run:
        """Create test run in Qase."""
        response = RunsApi(self._client).create_run(
            code=self._project_code,
            run_create=run_data,
        )
        created_run = cast(IdResponseAllOfResult, response.result)

        return self.get_run(cast(int, created_run.id))

    def load_cases_ids(
        self,
    ) -> list[int]:
        """Load all cases ids of project."""
        limit = 100
        cases: list[int] = []
        response = CasesApi(self._client).get_cases(
            code=self._project_code,
            limit=limit,
            offset=len(cases),
        )
        while True:
            response = CasesApi(self._client).get_cases(
                code=self._project_code,
                limit=limit,
                offset=len(cases),
            )
            response_cases = getattr(response.result, "entities", [])
            new_cases = [case.id for case in response_cases]
            cases += new_cases
            if not new_cases:
                break
        return cases

    def report_test_results(
        self,
        run: Run,
        report_data: ResultCreate,
    ) -> tuple[str, ResultCreate]:
        """Report test results back to Qase."""
        result = (
            ResultsApi(self._client)
            .create_result(
                code=self._project_code,
                id=cast(int, run.id),
                result_create=report_data,
            )
            .result
        )
        assert result
        return cast(str, result.hash), cast(ResultCreate, report_data.status)
